;;;			      mew-pgp.el
;;;
;;;	      Copyright (C) 1994-1997  Kazuhiko Yamamoto
;;;
;;;		   This emacs lisp library conforms
;;;		GNU GENERAL PUBLIC LICENSE Version 2.
;;;
;;; Author:  Kazu Yamamoto <Kazu@Mew.org>
;;; Created: Aug 17, 1994
;;; Revised: Oct 25, 1997
;;;

(defconst mew-pgp-version "mew-pgp.el version 0.46")

(require 'mew)

(defvar mew-pgp-key-suffix   ".pka")
(defvar mew-pgp-ascii-suffix ".asc")

(setq mew-pgp-string nil)
(setq mew-pgp-running nil)
(setq mew-pgp-failure nil)

(setq mew-pgp-decrypt-msg nil)
(setq mew-pgp-sign-msg nil)

(defconst mew-pgp-key-begin "-----BEGIN PGP PUBLIC KEY BLOCK-----")
(defconst mew-pgp-key-end   "-----END PGP PUBLIC KEY BLOCK-----")

(defconst mew-pgp-err-pass    'mew-err-pass)
(defconst mew-pgp-err-pubring 'mew-err-pubring)
(defconst mew-pgp-err-secring 'mew-err-secring)
(defconst mew-pgp-err-pubkey  'mew-err-pubkey)
(defconst mew-pgp-err-seckey  'mew-err-seckey)
(defconst mew-pgp-err-seckey-or-secring 'mew-err-seckey-or-secring)
(defconst mew-pgp-err-other   'mew-err-other)

(defvar mew-pgp-result-pass    "Pass phrase is wrong.")
(defvar mew-pgp-result-pubring "No public keyring.")
(defvar mew-pgp-result-secring "No secret keyring.")
(defvar mew-pgp-result-pubkey  "No his/her public key.")
(defvar mew-pgp-result-seckey  "No your secret key.")
(defvar mew-pgp-result-seckey-or-secring 
  "Not decrypted: no secret keyring or no your secret key.")
(defvar mew-pgp-result-other   "PGP failed for some reasons.")

(defconst mew-pgp-msg-bad-pass   "No passphrase")
(defconst mew-pgp-msg-enter      "Enter")
(defconst mew-pgp-msg-enter-pass "Enter pass phrase: ")
(defconst mew-pgp-msg-reenter-pass "Enter pass phrase: ")
(defconst mew-pgp-msg-no-key     "Key matching")
(defconst mew-pgp-msg-no-keyring "Keyring file")
(defconst mew-pgp-msg-no-seckey-or-secring
  "You do not have the secret key")

(defvar mew-pgp-prompt-enter-pass   "Enter pass phrase: ")
(defvar mew-pgp-prompt-reenter-pass "Re-enter pass phrase: ")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; a password function
;;

(defun mew-read-passwd (prompt)
  (let ((pass)
	(c 0)
	(echo-keystrokes 0)
	(ociea cursor-in-echo-area)
	(inhibit-input-event-recording t))
    (unwind-protect
	(progn
	  (setq cursor-in-echo-area 1)
	  (while (and (/= c ?\r) (/= c ?\n) (/= c ?\e) (/= c 7)) ;; ^G
	    (message "%s%s"
		     prompt
		     (make-string (length pass) ?.))
	    (setq c (read-char))
	    (if (char-equal c ?\C-u)
		(setq pass "")
	      (if (and (/= c ?\b) (/= c ?\177))
		  (setq pass (concat pass (char-to-string c)))
		(if (> (length pass) 0)
		    (setq pass (substring pass 0 -1))))))
	  (setq cursor-in-echo-area -1)
	  )
      (setq cursor-in-echo-area ociea)
      nil)
    (message "")
    (sit-for 0)
    (substring pass 0 -1)
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; PGP verifying
;;

(defun mew-pgp-verify-check ()
  (let (ret)
    (goto-char (point-min))
    (if (not (re-search-forward 
	      "\n\\(.*\\) signature from user \\(.*\\)\\." nil t))
	;; this is verification, so the error is about public key
	(if (re-search-forward mew-pgp-msg-no-key nil t)
	    (setq ret mew-pgp-result-pubkey)
	  (goto-char (point-min))
	  (if (re-search-forward  mew-pgp-msg-no-keyring nil t)
	      (setq ret mew-pgp-result-pubring)))
      (setq ret (concat (mew-match 1) " PGP sign " (mew-match 2)))
      (if (re-search-forward "not certified with enough" nil t)
	  (setq ret (concat ret " MARGINAL"))
	(goto-char (point-min))
	(if (re-search-forward "not trusted" nil t)
	    (setq ret (concat ret " UNTRUSTED"))
	  (goto-char (point-min))
	  (if (re-search-forward "not certified with a" nil t)
	      ;; PGP uses "unknown" for validity internally, but
	      ;; prints "undefined" instead of "unknown".
	      (setq ret (concat ret " UNDEFINED"))
	    (setq ret (concat ret " COMPLETE"))))))
    ret
    ))

(defun mew-pgp-verify (file1 file2)
  (message "PGP verifying ... ")
  (save-excursion
    (mew-set-buffer-tmp)
    (call-process mew-prog-pgp nil t nil 
		  "+batchmode=on" "+language=en" file2 file1)
    (message "PGP verifying ... done")
    (mew-pgp-verify-check)
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; PGP encrypting
;;

(defun mew-pgp-encrypt-check ()
  (let (ret)
    (goto-char (point-min))
    (if (re-search-forward mew-pgp-msg-no-key nil t)
	(setq ret mew-pgp-result-pubkey)
      (goto-char (point-min))
      (if (re-search-forward  mew-pgp-msg-no-keyring nil t)
	  (setq ret mew-pgp-result-pubring)))
    ret
    ))

(defun mew-pgp-encrypt (file1 decrypters)
  (message "PGP encrypting ... ")
  (let (file2 file3) ;; not unique if makes temp here
    (if (not (file-exists-p mew-temp-dir))
	(mew-make-directory mew-temp-dir)) ;; just in case
    (setq file2 (make-temp-name mew-temp-file)) 
    (save-excursion
      (mew-set-buffer-tmp)
      (insert "Version: 1\n")
      (write-region (point-min) (point-max) file2  nil 'no-msg)
      (setq file3 (concat (make-temp-name mew-temp-file) mew-pgp-ascii-suffix))
      (mew-set-buffer-tmp)
      (apply (function call-process) 
	     mew-prog-pgp nil t nil
	     "-ea" "+language=en" "+batchmode=on" "+armorlines=0"
	     "-o" file3 file1 decrypters)
      (message "PGP encrypting ... done")
      (list file2 nil file3 nil (mew-pgp-encrypt-check)) ;; both ctes are 7bit
      )
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; PGP decrypting
;;

(defun mew-pgp-decrypt (file1 file2)
  ;; file1 is a key file. just ignore.
  ;; file2 is an encrypted file with PGP.
  (message "PGP decrypting ... ")
  (setq mew-pgp-running 'decrypting)
  (setq mew-pgp-string nil)
  (setq mew-pgp-decrypt-msg nil)
  (setq mew-pgp-failure nil)
  (let ((process-connection-type mew-connection-type2)
	file3 process verify)
    (if (not (file-exists-p mew-temp-dir))
	(mew-make-directory mew-temp-dir)) ;; just in case
    (setq file3 (make-temp-name mew-temp-file)) 
    (setq process (start-process "PGP decrypt"
				 (current-buffer)
				 mew-prog-pgp
				 "+language=en" "+batchmode=off"
				 "-o" file3 file2))
    (mew-set-process-cs process mew-cs-autoconv mew-cs-noconv)
    (set-process-filter process 'mew-pgp-process-filter1)
    (set-process-sentinel process 'mew-pgp-process-sentinel)
    ;; Wait for the termination of PGP.
    ;; Emacs doesn't provide synchronize mechanism with
    ;; an asynchronous process. So, take this way. 
    (while mew-pgp-running
	(if mew-xemacs-p
	    (accept-process-output)
	  (sit-for 1)
	  ;; accept-process-output or sleep-for is not enough
	  (discard-input))
	)
    (message "PGP decrypting ... done")
    (save-excursion
      (mew-set-buffer-tmp)
      (insert mew-pgp-string)
      (setq verify (mew-pgp-verify-check))
      )
    (if verify 
	(setq mew-pgp-decrypt-msg (concat mew-pgp-decrypt-msg "\n\t" verify)))
    (list file3 mew-pgp-decrypt-msg)
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; PGP signing
;;

(defun mew-pgp-sign (file1)
  (message "PGP signing ... ")
  (setq mew-pgp-running 'signing)
  (setq mew-pgp-string nil)
  (setq mew-pgp-sign-msg nil)
  (setq mew-pgp-failure nil)
  (let ((process-connection-type mew-connection-type2)
	file2 process)
    (if (not (file-exists-p mew-temp-dir))
	(mew-make-directory mew-temp-dir)) ;; just in case
    (setq file2 (concat (make-temp-name mew-temp-file) mew-pgp-ascii-suffix))
    ;; not perfectly unique but OK
    (setq process (start-process "PGP sign"
				 (current-buffer)
				 mew-prog-pgp
				 "+language=en" "+batchmode=off"
				 "-o" file2
				 "-sab" file1))
    (mew-set-process-cs process mew-cs-autoconv mew-cs-noconv)
    (set-process-filter process 'mew-pgp-process-filter1)
    (set-process-sentinel process 'mew-pgp-process-sentinel)
    ;; Wait for the termination of PGP.
    ;; Emacs doesn't provide synchronize mechanism with
    ;; an asynchronous process. So, take this way. 
    (while mew-pgp-running
	(if mew-xemacs-p
	    (accept-process-output)
	  (sit-for 1)
	  ;; accept-process-output or sleep-for is not enough
	  (discard-input))
	)
    (message "PGP signing ... done")
    (list file2 nil "pgp-md5" mew-pgp-sign-msg) ;; return
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; PGP process functions
;;

(defun mew-pgp-process-sentinel (process event)
  (let ((decrypted "PGP decrypted. ")
	(msg ""))
    (if (not mew-pgp-failure)
	(cond
	 ((equal mew-pgp-running 'decrypting)
	  (setq mew-pgp-decrypt-msg decrypted))
	 ((equal mew-pgp-running 'signing)
	  (setq mew-pgp-sign-msg nil))
	 )
      (cond
       ;; sign or decrypt
       ((equal mew-pgp-failure mew-pgp-err-pass)
	(setq msg mew-pgp-result-pass))
       ;; decrypt-then-verify
       ((equal mew-pgp-failure mew-pgp-err-pubring)
	(setq msg decrypted))
       ;; decrypt-then-vrify
       ((equal mew-pgp-failure mew-pgp-err-pubkey)
	(setq msg decrypted))
       ;; sign
       ((equal mew-pgp-failure mew-pgp-err-secring)
	(setq msg mew-pgp-result-secring))
       ;; sign
       ((equal mew-pgp-failure mew-pgp-err-seckey)
	(setq msg mew-pgp-result-seckey))
       ;; decrypt
       ((equal mew-pgp-failure mew-pgp-err-seckey-or-secring)
	(setq msg mew-pgp-result-seckey-or-secring))
       ;; other
       ((equal mew-pgp-failure mew-pgp-err-other)
	(setq msg mew-pgp-result-other))
       )
      (cond
       ((equal mew-pgp-running 'decrypting)
	(setq mew-pgp-decrypt-msg msg))
       ((equal mew-pgp-running 'signing)
	(setq mew-pgp-sign-msg msg))
       )
      )
    (setq mew-pgp-running nil)
    ))

(defun mew-pgp-process-filter1 (process string)
  ;; sign or decrypt, not verify
  (setq mew-pgp-string (concat mew-pgp-string string))
  (cond
   ;; no secret key or no secring for decrypt
   ((string-match mew-pgp-msg-no-seckey-or-secring string)
    (setq mew-pgp-failure mew-pgp-err-seckey-or-secring)
    (set-process-filter process 'mew-pgp-process-filter3))

   ;; no secring for sign
   ((string-match mew-pgp-msg-no-keyring string)
    (setq mew-pgp-failure mew-pgp-err-secring)
    ;; Enter secret key filename: 
    (process-send-string process "\n")
    (set-process-filter process 'mew-pgp-process-filter3))

   ;; no secret key for sign
   ((string-match mew-pgp-msg-no-key string)
    (setq mew-pgp-failure mew-pgp-err-seckey)
    ;; Enter secret key filename: 
    (process-send-string process "\n")
    (set-process-filter process 'mew-pgp-process-filter3))

   ;; pass phrase for sign or decrypt
   ((string-match mew-pgp-msg-enter-pass string)
    (process-send-string
     process
     (format "%s\n" (mew-read-passwd mew-pgp-prompt-enter-pass)))
    (set-process-filter process 'mew-pgp-process-filter2))

   ;; just in case
   ((string-match mew-pgp-msg-enter string)
    (setq mew-pgp-failure mew-pgp-err-other)
    ;; Enter XXX:
    (process-send-string process "\n")
    (set-process-filter process 'mew-pgp-process-filter3))
   )
  )

(defun mew-pgp-process-filter2 (process string)
  (setq mew-pgp-string (concat mew-pgp-string string))
  (cond
   ;; re-enter pass phrase
   ((string-match mew-pgp-msg-reenter-pass string)
    (setq mew-pgp-string nil)
    (process-send-string 
     process
     (format "%s\n" (mew-read-passwd mew-pgp-prompt-reenter-pass)))
    (set-process-filter process 'mew-pgp-process-filter2))

   ;; pass phrases were wrong three times
   ((string-match mew-pgp-msg-bad-pass string)
    (setq mew-pgp-failure mew-pgp-err-pass)
    (set-process-filter process 'mew-pgp-process-filter3))

   ;; no pubring for verify
   ((string-match mew-pgp-msg-no-keyring string)
    (setq mew-pgp-failure mew-pgp-err-pubring)
    ;; Enter public key filename: 
    (process-send-string process "\n")
    (set-process-filter process 'mew-pgp-process-filter3))

   ;; no public key for verify
   ((string-match mew-pgp-msg-no-key string)
    (setq mew-pgp-failure mew-pgp-err-pubkey)
    ;; Enter public key filename: 
    (process-send-string process "\n")
    (set-process-filter process 'mew-pgp-process-filter3))

   ;; just in case
   ((string-match mew-pgp-msg-enter string)
    (setq mew-pgp-failure mew-pgp-err-other)
    ;; Enter XXX:
    (process-send-string process "\n")
    (set-process-filter process 'mew-pgp-process-filter3))
   )
  )

(defun mew-pgp-process-filter3 (process string)
  ;; ending or error
  (setq mew-pgp-string (concat mew-pgp-string string))
  (cond
   ;; just in case
   ((string-match mew-pgp-msg-enter string)
    (setq mew-pgp-failure mew-pgp-err-other)
    ;; Enter XXX:
    (process-send-string process "\n")
    (set-process-filter process 'mew-pgp-process-filter3))
   )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; shortcut methods
;;

(defun mew-pgp-sign-letter ()
  "Sign the single part draft with PGP. Input your passphrase."
  (interactive)
  (mew-pgp-encode-letter (list (list mew-ct-mls mew-ct-pgs)))
  )

(defun mew-pgp-encrypt-letter ()
  "Encrypt the single part draft with PGP."
  (interactive)
  (mew-pgp-encode-letter (list (list mew-ct-mle mew-ct-pge)))
  )

(defun mew-pgp-sign-encrypt-letter ()
  "Sign then encrypt the single part draft with PGP. Input your passphrase."
  (interactive)
  (mew-pgp-encode-letter (list (list mew-ct-mls mew-ct-pgs)
			       (list mew-ct-mle mew-ct-pge)))
  )

(defun mew-pgp-encode-letter (privacy)
  (cond
   ((null (mew-which mew-prog-pgp exec-path))
    (message "PGP doesn't exist"))
   ((mew-header-get-value "Content-Type:")
    (message "You can't sign/encrypt this draft due to Content-Type:"))
   ((mew-attach-p)
    (message "You can't sign/encrypt this draft with the shortcut method"))
   (t
    (mew-draft-make-backup 'single)
    (goto-char (marker-position mew-draft-buffer-header))
    ;; the beginning of "----"
    (let ((beg (point)) decrypters)
      (forward-line)
      ;; cursor is just after "----\n"
      (delete-region beg (point))
      ;; same as mew-encode-decrypters-list
      (setq decrypters (mew-header-decrypters-collect '("To:" "Cc:")))
      (mew-encode-singlepart
       (mew-encode-syntax-single "text-file" nil nil nil privacy decrypters)
       nil nil t)
      (mew-draft-make-header)
      (goto-char (point-min))
      (sit-for 0) ;; to display the top of buffer 
      (re-search-forward mew-eoh2)
      (beginning-of-line)
      (setq mew-draft-buffer-header (point-marker)) ;; just in case
      ))
   ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; key distribution
;;

(defun mew-attach-pgp-public-key ()
  "Extract the PGP key for the inputed user on \".\". in attachments"
  (interactive)
  (if (not (mew-attach-not-line012-1))
      (message "Can't link here.")
    (let* ((error nil)
	   (nums (mew-attach-nums))
	   (subdir (mew-attach-expand-path mew-encode-syntax nums))
	   (mimedir (mew-expand-folder (mew-draft-to-mime (buffer-name))))
	   user file filepath begin end)
      ;; mimedir / {subdir/} dir
      (if (not (equal subdir "")) 
	  (setq mimedir (expand-file-name subdir mimedir)))
      ;; mimedir / file
      (setq filepath (mew-random-filename mimedir mew-pgp-key-suffix))
      (if (null filepath)
	  (message "Could not make a file for pgp key, sorry.")
	(setq file (file-name-nondirectory filepath))
	(setq user (mew-input-address2 "Who's key? (%s): " mew-mail-address))
	(save-excursion
	  (mew-set-buffer-tmp)
	  (call-process mew-prog-pgp nil t nil "-fkxa" user)
	  (goto-char (point-min))
	  (if (not (re-search-forward mew-pgp-key-begin nil t))
	      (setq error t)
	    (beginning-of-line)
	    (setq begin (point))
	    (if (not (re-search-forward mew-pgp-key-end nil t))
		(setq error t)
	      (beginning-of-line)
	      (forward-line)
	      (setq end (point)))
	    (write-region begin end filepath nil 'no-msg)))
	(if error
	    (message "can't extract pgp key for %s" user)
	  (setq mew-encode-syntax
		(mew-syntax-insert-entry
		 mew-encode-syntax 
		 nums
		 (mew-encode-syntax-single file mew-type-apk nil user)))
	  (mew-encode-syntax-print mew-encode-syntax))
	))
    ))

(defvar mew-pgp-tmp-file nil)

(defun mew-mime-pgp-keys (begin end &optional params)
  "A function to add PGP keys in Application/PGP-Keys to your 
public keyring."
  (interactive)
  (if (not (mew-which mew-prog-pgp exec-path))
      (message "PGP is not found")
    (if (not (mew-y-or-n-p "Add this PGP key onto your public keyring? "))
	()
      (if (not (file-exists-p mew-temp-dir))
	  (mew-make-directory mew-temp-dir));; just in case
      (setq mew-pgp-tmp-file (make-temp-name mew-temp-file))
      (save-excursion
	(set-buffer (mew-current-get 'cache))
	(mew-frwlet mew-cs-noconv mew-cs-autoconv
	 (write-region begin end mew-pgp-tmp-file nil 'no-msg)
	 )
	(set-buffer (mew-buffer-message))
	(let ((buffer-read-only nil))
	  (message "Adding PGP keys ... ")
	  (call-process mew-prog-pgp nil t nil 
			"+batchmode=on" "-ka" mew-pgp-tmp-file)
	  (message "Adding PGP keys ... done")
	  (insert "\n\n"
		  "**************** IMPORTANT NOTE ****************\n"
		  "When Mew adds PGP keys onto your public keyring,\n"
		  "it is careless about both TRUST and VALIDITY.\n"
		  "It is YOU who set these values. Please use\n"
		  "\"pgp -ke\" and \"pgp -ks\" to change them.\n"
		  "If you don't know what TRUST and VALIDITY is,\n"
		  "you should learn the web of trust system BEFORE\n"
		  "using PGP to protect your privacy.\n"
		  "**************** IMPORTANT NOTE ****************\n"
		  )
	  ))
      (if (file-exists-p mew-pgp-tmp-file)
	  (delete-file mew-pgp-tmp-file))
      )
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; <Decrypt> or <Decrypt-then-Verify>
;; 
;; no seckey no secring:: <1>
;; You do not have the secret key needed to decrypt this file.
;; 
;;    <Verify> <2>
;; 	Keyring file 'pubring.pgp' does not exist. 
;; 	Enter public key filename:
;; 
;; 	Key matching expected Key ID 1B8BF431 not found in file 'pubring.pgp'.
;; 	Enter public key filename:
;; 
;; <Sign> <1>
;; 
;; A secret key is required to make a signature. 
;; Keyring file 'secring.pgp' does not exist. 
;; Enter secret key filename: 
;; 
;; A secret key is required to make a signature. 
;; Key matching userid 'hoge' not found in file 'secring.pgp'.
;; Enter secret key filename: 
;; 
;; <Encrypt>
;; 
;; Keyring file 'pubring.pgp' does not exist. 
;; 
;; Key matching userid 'hoge' not found in file 'pubring.pgp'.
;; 
;; <Verify>
;; Keyring file 'pubring.pgp' does not exist. 
;; 
;; Key matching expected Key ID 1B8BF431 not found in file 'pubring.pgp'.

(provide 'mew-pgp)
