;;;		mew.el :: Messaging in the Emacs World
;;;
;;;	      Copyright (C) 1994-1997  Kazuhiko Yamamoto
;;;
;;;		   This emacs lisp library conforms
;;;		GNU GENERAL PUBLIC LICENSE Version 2.
;;;
;;; Author:  Kazu Yamamoto <Kazu@Mew.org>
;;; Created: Mar 22, 1994
;;; Revised: Oct 25, 1997
;;;
;;; The updated version is available from:
;;;	ftp://ftp.aist-nara.ac.jp/pub/elisp/Mew/mew-current.tar.gz
;;;	http://www.mew.org/
;;;
;;; Minimum setup:
;;;	(autoload 'mew "mew" nil t)
;;;	(autoload 'mew-send "mew" nil t)
;;;	(setq mew-mail-domain-list '("your mail domain"))
;;;	(setq mew-icon-directory "icon directory")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Mew version
;;;

(defconst mew-version "Mew version 1.92")
(provide 'mew)
(require 'mew-env)
(require 'mew-func)

(defun mew-version-show ()
  "Show mew-version in minibuffer."
  (interactive)
  (message "%s" mew-version))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; For developers
;;;

(defvar mew-debug nil)
;(setq mew-debug nil)
;(setq mew-debug t)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; User customizable variables. See also other ".el" files.
;;;

;;;
;;; Common
;;;

(defvar mew-mail-domain-list nil
  "*Your e-mail address domain list like 
(\"is.aist-nara.ac.jp\" \"csce.kyushu-u.ac.jp\").
They are used for mail-domain completion in draft mode(C-cTAB)")

(defvar mew-mail-domain (or (car mew-mail-domain-list) "")
  "*Your e-mail address domain.")

(defvar mew-mail-address-list nil
  "*The addresses included in this list never appear on the Cc:
field on a draft buffer.")

(cond 
 ((string= mew-mail-domain "")
  (defvar mew-mail-address (user-login-name)))
 (t
  (defvar mew-mail-address (concat (user-login-name) "@" mew-mail-domain)))
 )

(defvar mew-window-configuration 
  '((summary (1  0))
    (message (8 31))
    (draft   (1  0)))
  "*Ratio of windows")

;;;
;;; Summary mode
;;;

(defvar mew-analysis t
  "*MIME syntax analysis and cache when non-nil.
   Skip syntax analysis when nil.")

(defvar mew-cache-size 10
  "*Number of buffer for message cache.")

(defvar mew-summary-cache-file ".mew-cache"
  "*Cache file for summary mode contents.")

(defvar mew-summary-cache-use t
  "*If non-nil, insert mew-summary-cache-file into summary mode buffer 
when the summary mode is visited and it does not exist in buffer. 
mew-summary-cache-file is automatically saved on inc and scan.")

(defvar mew-window-use-full nil
  "*Dynamic window configuration when nil. This variable will be obsolated
because mew will support mew-window-percent(e.g 50, 100, etc.).")

(defvar mew-window-home-buffer "*scratch*"
  "*Buffer name to return if window stack is empty.")

(defvar mew-summary-show-direction 'down
  "*Direction for SPC at end of message. 
'up 'down 'next(current direction) 'stop")

(defvar mew-mime-compose-folder-delete 'delete
 "*Action to delete MIME draft directory after sending a multipart letter.
Proper value is 'ask', 'delete', and 'retain'.")

(defvar mew-summary-trace-directory t
  "*If non-nil, trace directory when mew-summary-goto-folder.")

(defvar mew-summary-recenter-p t)
(defvar mew-break-pages t)
(defvar mew-page-delimiter "^\^L")

(defvar mew-bold-url-size 3000
  "*Use bold font at url maximize size.
If nil, all contents are parsed.")

;;;
;;; Draft mode
;;;

(defvar mew-header-separator "----")
(defvar mew-eoh "^$")
(defvar mew-eoh2 (format "^\\(%s\\|\\)$" (regexp-quote mew-header-separator)))
(defvar mew-lwsp "^[ \t]")

(defvar mew-reply-to nil
  "*A value inserted into Reply-To: field in draft mode if non-nil.")

(defvar mew-fcc nil
  "*A value inserted into Fcc: field in draft mode if non-nil.")

(defvar mew-cc nil
  "*A value inserted into Cc: field in draft mode if non-nil.")

(defvar mew-dcc nil
  "*A value inserted into Dcc: field in draft mode if non-nil.")

(defvar mew-from nil
  "*A value inserted into From: field in draft mode if non-nil.")

(defvar mew-header-alist nil
  "*Alist of key and value pair for header field to be inserted on draft.
'((\"X-fingerprint:\" . \"6B 63 38 88 67 5E 96 8E  CE A4 62 73 3F 11 64 94\")
  (\"X-URL:\" . \"http://fukuda.aist-nara.ac.jp/~kazu/\"))"
  )

(defvar mew-auto-add-content-type nil
  "*Automatically adds MIME Content-Type: header if non-nil, asks otherwise.")

(defvar mew-ask-cc t
  "*Prompt user for CC: field if non-nil")

(defvar mew-ask-subject nil
  "*Prompt user for Subject: field when send if non-nil")

(defvar mew-ask-range t
  "*If non-nil, ask for a range for scanning. If false, always
use default.")

(defvar mew-ask-newsgroups nil
  "*If non-nil, prompt user if he/she want to include Newsgroups: field")

;;;
;;; Signature
;;; 

(defvar mew-signature-file "~/.signature"
  "*A signature file to be inserted in Draft mode. To support multiple
signature files, use \"c-sig.el\".")

(defvar mew-signature-insert-last nil 
  "*If non-nil, the signature file is inserted in the last of body. 
Otherwise, it is inserted at the currect point. If you created multipart
and mew-signature-as-lastpart is non-nil, this variable is ignored.")

(defvar mew-signature-as-lastpart nil
  "*If non-nil, the signature file is appended as the final part 
when you create multipart.")

(defvar mew-signature-description "My signature"
  "*This variable is used as a description if the signature is appended
as the final part.")

;;;
;;; Summary and Draft glue
;;;

;;
;; Temporary solution to decide To: when reply
;;

(defvar mew-noreplyto-to-list '("From:")
  "*To: field list picked at mew-summary-reply if Reply-To: doesn't exist.")

(defvar mew-noreplyto-cc-list '("To:" "Cc:" "Apparently-To:")
  "Cc: field list picked at mew-summary-reply if Reply-To: doesn't exist.")

(defvar mew-replyto-to-list '("Reply-To:" "From:")
  "*To: field list picked at mew-summary-reply if Reply-To: exists.")

(defvar mew-replyto-cc-list '("To:" "Cc:" "Apparently-To:")
  "*Cc: field list picked at mew-summary-reply if Reply-To: exists.")

;;;
;;; Misc
;;;

(defvar mew-modeline-id '("Mew: %12b")
  "*A default value of modeline-buffer-identification for each Mew mode.")

;;BBDB not fully supported
(defvar mew-use-bbdb nil
  "*Use Insidious Big Brother Database if t")

;; Icon
(defvar mew-multipart-icon-position 'right
  "*Position where multipart icons are displayed. 
If 'left, displayed at the left size of the default toolbar.
If 'right, displayed at the right size of the default toolbar.
Otherwise, not displayed.")

;;
;; X face
;;
(defvar mew-x-face-filter '("uncompface" "ikon2xbm"))
(defvar mew-x-face-prog "xv")
(defvar mew-x-face-args nil)
(defvar mew-x-face-file "~/.xface"
  "*If non-nil and the file exists, X-Face: fields is inserted.")

;;
;; Petname
;;

(defvar mew-use-petname nil)

;;;
;;; File
;;;

(defvar mew-use-folders-file-p nil
  "If non-nil, Mew try to load mew-folders-file to make bootup faster.
And C-u Z saves folder list to mew-folders-file.")

(defvar mew-folders-file ".folders"
  "*A file to save folder list. It is used on start-up to
make start-up much faster.")
(defvar mew-folders-ignore '("+from")
  "*Folder list to contain messages whose To: is you and From: is
personal, e.g. a friend. Directories under this folder is ignored
for refile guess.")
(defvar mew-folders-default-folder "+from"
  "*A default folder used by mew-guess-by-default.
mew-guess-by-default append the \"From:\" value to it.
e.g. \"+from/kazu\"")

(defvar mew-file-append-p nil
  "*If non-nil, a message or a part is appended to the existing file
on mew-summary-save(\"y\"). Otherwise overwrited.")

(defvar mew-temp-file-initial
  (expand-file-name (user-login-name) (or (getenv "TMP") "/tmp"))
  "*Hint to make a secure directory on the local file system.
On setup phase Mew make a secure directory from this variable
and set mew-temp-file a file name prefex contained the directory name.
The directory must be unreadable from others, otherwise it might become 
a big security hole. Mew never uses 'call-process-region' rather does use
'call-process' creating a temporary file with mew-temp-file by itself.
If 'call-process-region' is used, Emacs creates a temporary file
(probably in /tmp). So bad guys can wiretap the temporary file.")

(defvar mew-path-separator "/")
(defconst mew-home (file-name-as-directory "~"))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Hooks
;;;

(defvar mew-init-hook nil
  "*Hook called at initialize time.")
(defvar mew-summary-mode-hook nil
  "*Hook called in summary mode.")
(defvar mew-virtual-mode-hook nil
  "*Hook called in virtual mode.")
(defvar mew-draft-mode-hook nil
  "*Hook called in draft mode.")
(defvar mew-message-mode-hook nil
  "*Hook called in message mode.")
(defvar mew-message-hook nil
  "*Hook called whenever message displayed.")
(defvar mew-send-hook nil
  "*Hook called before sending a message")
(defvar mew-suspend-hook nil
  "*Hook called on suspend.")
(defvar mew-quit-hook nil
  "*Hook called on quit.")
(defvar mew-summary-inc-sentinel-hook nil
  "*Hook called when inc finished.")
(defvar mew-summary-scan-sentinel-hook nil
  "*Hook called when scan finished.")
(defvar mew-summary-exec-hook nil
  "*Hook called when mew-summary-exec finished.")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Buffers
;;;

(defconst mew-buffer-message     "*Mew message*")
(defconst mew-buffer-mime       " *mew mime*")
(defconst mew-buffer-cache      " *mew cache*")
(defconst mew-buffer-watch       "*Mew watch*")
(defconst mew-buffer-tmp        " *mew tmp*")
(defconst mew-buffer-burst      " *mew burst*")
(defconst mew-buffer-completions "*Mew completions*")
(defconst mew-buffer-whom        "*Mew whom*")

(defun mew-buffer-message ()
  (if window-system
      (concat
       mew-buffer-message
       (int-to-string
	(mew-member-case-equal 
	 (cdr (assq
	       'window-id
	       (frame-parameters (window-frame (selected-window)))))
	 (sort
	  (mapcar
	   (function (lambda (frame)
		       (cdr (assq
			     'window-id
			     (frame-parameters frame)))))
	   (frame-list))
	  (function string<)))))
    mew-buffer-message
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Buffer local variables
;;;

(defvar mew-summary-buffer-start-point nil)
(defvar mew-summary-buffer-process nil)
(defvar mew-summary-buffer-string nil)
(defvar mew-summary-buffer-direction nil)
(defvar mew-summary-buffer-folder-cache-time nil)
(defvar mew-summary-buffer-refile nil)
(defvar mew-summary-buffer-disp-msg t)
(defvar mew-summary-buffer-left-msgs "-")
(defvar mew-watch-buffer-process nil)
(defvar mew-draft-buffer-header nil)
(defvar mew-draft-buffer-attach nil)
(defvar mew-cache-attribute nil)
(defvar mew-decode-syntax nil)
(defvar mew-decode-error nil)
(defvar mew-encode-syntax nil)
(defvar mew-property-cursor-line nil)

(mapcar
 (function make-variable-buffer-local)
 (list 'mew-summary-buffer-start-point
       'mew-summary-buffer-process
       'mew-summary-buffer-string
       'mew-summary-buffer-direction
       'mew-summary-buffer-folder-cache-time
       'mew-summary-buffer-refile
       'mew-summary-buffer-disp-msg
       'mew-summary-buffer-left-msgs
       'mew-watch-buffer-process
       'mew-draft-buffer-header
       'mew-draft-buffer-attach
       'mew-cache-attribute
       'mew-encode-syntax
       'mew-decode-syntax
       'mew-decode-error
       'mew-property-cursor-line
       ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Sub-programs
;;;

(defvar mew-prog-imget       "imget")
(defvar mew-prog-imls        "imls")
(defvar mew-prog-imcat       "imcat")
(defvar mew-prog-imali       "imali")
(defvar mew-prog-impath      "impath")
(defvar mew-prog-impack      "impack")
(defvar mew-prog-immv        "immv")
(defvar mew-prog-imrm        "imrm")
(defvar mew-prog-imput       "imput")
(defvar mew-prog-imsort      "imsort")
(defvar mew-prog-imstore     "imstore")
(defvar mew-prog-imgrep      "imgrep")
(defvar mew-prog-mime-encode "mewencode")
(defvar mew-prog-mime-decode "mewdecode")
(defvar mew-prog-audio       (list "cat" "/dev/audio" mew-buffer-tmp nil))
;(defvar mew-prog-audio      '("arecord" nil nil nil "-silentlevel" "-100"))
(defvar mew-prog-unshar      "unshar")
(defvar mew-prog-uumerge     "uumerge")
(defvar mew-prog-tar         "tar")
(defvar mew-prog-compress    "compress")
(defvar mew-prog-gzip        "gzip")
(defvar mew-prog-utime       "utime")
(defvar mew-prog-pgp         "pgp")
(defvar mew-prog-shell       "/bin/sh")
;(defvar mew-prog-shell       "cmd.exe")

(defvar mew-prog-shell-arg   "-c")

(defvar mew-prog-imget-arg-list nil
  "*Argument list for mew-prog-imget")
(defvar mew-prog-imls-arg-list nil
  "*Argument list for mew-prog-imls")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Initialize global variables
;;;

(defvar mew-mail-path nil)
(defvar mew-news-path nil)
(defvar mew-inbox-folder nil)
(defvar mew-draft-folder nil)
(defvar mew-trash-folder nil)

(defvar mew-msg-rm-policy 'trashonly
  "Set remove policy. You can set one of the followings:
'totrash   : Refile to the +trash folder if not in the +trash folder.
             Just umark the 'D' mark if in the +trash folder.
'always    : Really remove messages marked with 'D' always anyway.
'trashonly : Really remove messages marked with 'D' if in the +trash folder.
             In other fonders, refile to the +trash folder.
'uselist   : Really remove messages marked with 'D' if in a folder found
	     in @samp{mew-msg-rm-folder-list}. In other folders, refile 
             to the +trash folder.
otherwise  : considered as 'totrash.
")

(defvar mew-msg-rm-folder-list nil
  "Folder list to remove message really.")

(defvar mew-petname-file nil)
(defvar mew-draft-mime-folder nil
  "A directory where attachments are contained.")
(defvar mew-config-alist
  '(("MailPath" . mew-mail-path)
    ("NewsPath" . mew-news-path)
    ("InboxFolder" . mew-inbox-folder)
    ("DraftFolder" . mew-draft-folder)
    ("TrashFolder" . mew-trash-folder)
    ("PetnameFile" . mew-petname-file))
  )

(setplist 'mew-current-state 
	  (list
	   'message    nil 
	   'cache      nil
	   'part       nil
	   'window     nil
	   ))

(defun mew-current-get (key)
  (get 'mew-current-state key))

(defun mew-current-set (key value)
  (put 'mew-current-state key value))

(setq mew-alias-alist  ())
(setq mew-folder-list  ())
(setq mew-folder-alist ())
(setq mew-petname-alist ())
(setq mew-clean-up-buffers ())

(setq mew-message-citation nil)
(setq mew-message-citation-buffer nil) ;; should be nil

(setq mew-process-file-alist nil)

;; See mew-temp-file-initial.
(setq mew-temp-dir nil)  ;; the default is "/tmp/user_name_uniq"
(setq mew-temp-file nil) ;; the default is "/tmp/user_name_uniq/mew"

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Window configuration stack
;;;

(defvar mew-window-stack nil)

(defun mew-window-push ()
  (let ((frame (selected-frame))
	(config (current-window-configuration)))
    (setq mew-window-stack (cons (cons frame config) mew-window-stack))
    )
  )

(defun mew-window-pop ()
  (let* ((frame (selected-frame))
	 (assoc (assoc frame mew-window-stack)))
    (if (and assoc (window-configuration-p (cdr assoc)))
	(set-window-configuration (cdr assoc))
      (switch-to-buffer (get-buffer-create mew-window-home-buffer)))
    (setq mew-window-stack (delete assoc mew-window-stack))
    )
  )

;;;
;;; Window configuration
;;;

(defun mew-window-configure (nbuf action)
;;; action : summary, message, draft or list
;;; list for action (1 0)  for Summary only
;;; list for action (3 10) for Summary and Message
  (if (equal action 'summary)
      (mew-current-set 'message nil))
  (let* ((windows
	  (if (listp action) 
	      action
	    (car (cdr (assq action mew-window-configuration)))))
	 (obufwin (get-buffer-window (current-buffer)))
	 (msgwin  (get-buffer-window (mew-buffer-message)))
	 (height nil) (winsum nil) (sum-height 0) (msg-height 0))
    (setq height (+ (if obufwin (window-height obufwin) 0)
		    (if msgwin  (window-height msgwin)  0)))
    (if (or mew-window-use-full
	    (<= height (* 2 window-min-height)))
	(progn
	 ;; Delete other windows and use full emacs window.
	 (delete-other-windows)
	 (setq height (window-height (selected-window)))))
    (if (get-buffer (mew-buffer-message))
	(delete-windows-on (mew-buffer-message))
      (save-excursion
	(set-buffer (get-buffer-create (mew-buffer-message)))
	;; "truncate?" is asked in Message mode.
	;; so set the same toolbar as Sumamry mode
	(if mew-icon-p
	    (set-specifier default-toolbar
			   (cons (current-buffer) mew-summary-toolbar)))
	(mew-message-mode)))
    (setq winsum (apply (function +) windows))
    (if (not (zerop (nth 0 windows)))
	(setq sum-height (max window-min-height
			     (/ (* height (nth 0 windows)) winsum))))
    (if (and (equal action 'summary) (equal (% sum-height 2) 1)) 
	(setq sum-height (1+ sum-height)))
    (if (not (zerop (nth 1 windows)))
	(setq msg-height (max window-min-height
			     (- height sum-height))))
    (setq height (+ sum-height msg-height))
    (if (null (zerop sum-height))
	(switch-to-buffer nbuf 'norecord))
    (if (zerop msg-height)
	()
      (split-window nil sum-height)
      (other-window 1)
      (switch-to-buffer (mew-buffer-message) 'norecord))
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Bootstrap
;;;

(defun mew (&optional arg)
  (interactive "P")
  (mew-window-push)
  (if (null mew-mail-path) (mew-init))
  (if arg
      (mew-summary-goto-folder) ;; C-u
    (mew-summary-get)
    )
  )

(defun mew-send (&optional to cc subject)
  (interactive)
  (mew-current-set 'window (current-window-configuration))
  (if (null mew-mail-path) (mew-init))
  (mew-summary-send t to cc subject)
  )

;;;
;;; Functions for boot time
;;;

(defun mew-init ()
  (mew-hello)
  (message "Setting Mew world ...")
  (mew-set-environment)
  (if mew-use-bbdb (require 'bbdb-com)) ;; bbdb is implicitly required
  (run-hooks 'mew-init-hook)
  (message "Setting Mew world ... done")
  )

(defun mew-set-environment ()
  (let (error-message)
    (condition-case nil
	(progn
	  (if (string-match "^18" emacs-version)
	      (progn
		(setq error-message "Not support Emacs 18 nor Mule 1\n")
		(error "")
		))
	  ;;
	  ;; initialize IM variables
	  ;;
	  (if (setq error-message (mew-read-config))
	      (error ""))
	  ;;
	  (setq mew-clean-up-buffers (list mew-inbox-folder)) ; for quiting
	  ;;
	  ;; mew-temp-file must be local and readable for the user only
	  ;; for privacy/speed reasons.
	  ;;
	  (setq mew-temp-dir (make-temp-name mew-temp-file-initial))
	  (mew-make-directory mew-temp-dir)
	  (set-file-modes mew-temp-dir 448) ;; decimal for octal 700
	  (setq mew-temp-file (expand-file-name "mew" mew-temp-dir))
	  ;;
	  (mew-status-update t)
	  ;;
	  (add-hook 'kill-emacs-hook (function mew-mark-clean-up-all))
	  (add-hook 'kill-emacs-hook (function mew-temp-dir-clean-up))
	  )
      (error
       (set-buffer mew-buffer-hello)
       (goto-char (point-max))
       (insert "\n\nMew errors:\n\n")
       (and error-message (insert error-message))
       (set-buffer-modified-p nil)
       (setq buffer-read-only t)
       (error "Mew found some errors above.")) ;; cause an error again
      )
    ))

(defun mew-status-update (arg)
  "Update the list of aliases. If you type 'C-u Z' the list of folders
are also updated in addition to that of aliases.  If 
'mew-use-folders-file-p' is t, the list of folders is stored in
\"~/Mail/.folders\"."
  (interactive "P")
  (cond
   (arg ;; C-u
    (message "Updating aliases and folders ... ")
    (setq mew-alias-alist (mew-alias-make-alist))
    (setq mew-petname-alist (mew-petname-make-alist))
    (setq mew-folder-list (mew-folder-make-list (interactive-p)))
    (setq mew-folder-alist (mew-folder-make-alist mew-folder-list))
    ;;
    (message "Updating aliases and folders ... done")
    )
   (t
    (message "Updating aliases ... ")
    (setq mew-alias-alist (mew-alias-make-alist))
    (setq mew-petname-alist (mew-petname-make-alist))
    (message "Updating aliases ... done"))
   )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Config
;;;

(defun mew-read-config ()
  (mew-set-buffer-tmp)
  (if (not (mew-which mew-prog-impath exec-path))
      "impath is not found in 'exec-path'"
    (call-process mew-prog-impath nil t nil "--path=yes")
    (goto-char (point-min))
    (let (key value pair)
      (while (not (eobp))
	(if (looking-at "^\\(.*\\)=\\(.*\\)$")
	    (progn
	      (setq key (mew-match 1))
	      (setq value (mew-match 2))
	      (if (setq pair (mew-assoc-match2 key mew-config-alist 0))
		  (set (cdr pair) value))
	      ))
	(forward-line)
	))
    (let (errmsg)
      (if mew-mail-path
	  ()
	(setq errmsg (concat errmsg "\tMailDir was NOT found in Config.\n")))
      (if mew-news-path
	  ()
	(setq errmsg (concat errmsg "\tNewsDir was NOT found in Config.\n")))
      (if mew-inbox-folder
	  ()
	(setq errmsg (concat errmsg "\tInboxFolder was NOT found in Config.\n")))
      (if mew-draft-folder
	  ;; xxx ugh! what a hard coding... +draft/mime
	  (setq mew-draft-mime-folder
		(concat (file-name-as-directory mew-draft-folder) "mime"))
	(setq errmsg (concat errmsg "\tDraftFolder was NOT found in Config.\n")))
      (if mew-trash-folder
	  ()
	(setq errmsg (concat errmsg "\tTrashDir was NOT found in Config.\n")))
      (if errmsg
	  errmsg ;; return value
	(let ((folders (list mew-inbox-folder mew-draft-folder mew-trash-folder))
	      target)
	  (while folders
	    (setq target (mew-expand-folder (car folders)))
	    (if (file-exists-p target)
		()
	      (make-directory target)
	      (message "%s was created" target))
	    (setq folders (cdr folders))
	    ))
	nil) ;; return value
      )))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; e-mail address alias
;;;

(defun mew-alias-make-alist ()
  "Make alias alist with association of (alias . expantion) from
IM \"imali\" command. Currently, only \"user: user@domain\" syntax
is supported."
  (save-excursion
    (let ((case-fold-search t)
	  (alias nil)
	  (expn nil)
	  (alist nil))
      (mew-set-buffer-tmp)
      (call-process mew-prog-imali nil t nil)
      ;; concat separated lines by comma
      (goto-char (point-min))
      (while (re-search-forward ",$" nil t)
	(end-of-line)
	(forward-char 1)
 	(delete-backward-char 1))
      ;;
      (goto-char (point-min))
      (while (re-search-forward "^\\([^:]+\\):[ \t]*\\(.*\\)$" nil t)
	(setq alias (mew-match 1)
	      expn (mew-match 2))
	;; append for first assoc comes first
	(setq alist (cons (cons alias expn) alist)))
      ;; load from mew-refile-from-alist
      (setq alist (mew-refile-alist-append-alias alist))
      (nreverse alist) ; return value
      )))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Folders
;;;

(defun mew-folder-make-list (updatep)
  (save-excursion
    (let ((case-fold-search t)
	  (folders ())
	  (folder nil)
	  (start nil)
	  (file (expand-file-name mew-folders-file mew-mail-path)))
      (mew-set-buffer-tmp)
      (if (and (not updatep)
	       mew-use-folders-file-p
	       (file-exists-p file))
	  (insert-file-contents file)
	(call-process mew-prog-impath nil t nil "--list=yes" "--src=+")
	(call-process mew-prog-impath nil t nil "--list=yes" "--src==")
	(if mew-use-folders-file-p
	    (write-region (point-min) (point-max) file nil 'no-msg))
	)
      (goto-char (point-min))
      (while (not (eobp))
        (setq start (point))
	(if (not (looking-at "[+=]"))
	    (forward-line)
	  (forward-line)
	  (setq folder (mew-buffer-substring start (1- (point))))
	  (if (and (car folders)
		   (string-match (concat "^" (regexp-quote 
					      (file-name-as-directory
					       (car folders))))
				 folder))
	      ;; regexp-quote is not necessary since not "+".
	      (setq folders 
		    (cons folder 
			  (cons (file-name-as-directory (car folders))
				(cdr folders))))
	    (setq folders (cons folder folders))))
	)
      folders ;; return value
      )))

(defmacro mew-folder-make-alist (list)
  (` (mapcar (function mew-folder-pair) (, list)))
  )

(defun mew-folder-pair (folder)
  (let* ((dir (directory-file-name (mew-folder-to-dir folder)))
	 ;; foo/bar  -> foo/bar
	 ;; foo/bar/ -> foo/bar
	 (subdir (file-name-nondirectory dir)))
	 ;; foo/bar -> bar 
	 ;; foo -> foo
    (if (mew-folders-ignore-p folder)
	(list folder nil)
      (list folder subdir)
      )
    ))

(defun mew-folders-ignore-p (folder)
  (let ((ignores mew-folders-ignore))
    (catch 'ignore
      ;; while always returns nil
      (while ignores
	(if (string-match (concat "^" (car ignores)) folder)
	    (throw 'ignore t))
	(setq ignores (cdr ignores))
	))
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Petnames
;;;

(defun mew-petname-make-alist ()
  (if (and mew-use-petname mew-petname-file (file-exists-p mew-petname-file))
      (save-excursion
	(let (ret)
	  (mew-set-buffer-tmp)
	  (insert-file-contents mew-petname-file)
	  (goto-char (point-min))
	  (while (not (eobp))
	    (if (looking-at "^\\([^ \t]+\\)[ \t]+\"\\(.*\\)\"$")
		(setq ret (cons (cons (mew-match 1) (mew-match 2)) ret)))
	    (forward-line))
	  ret))
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Teer down
;;;

(defun mew-temp-dir-clean-up ()
  "A function to remove Mew's temporary directory recursively. 
It is typically called by kill-emacs-hook."
  (if (and mew-temp-dir (file-exists-p mew-temp-dir))
      (mew-delete-directory-recursively mew-temp-dir))
  )

(defun mew-summary-suspend ()
  "Suspend Mew then switch to another buffer. All buffers of 
Mew retain, so you can resume with buffer operations."
  (interactive)
  (mew-window-pop)
  ;; bury buffers
  (if window-system
      (let ((x (buffer-list))
	    (regexp (concat "^" (regexp-quote mew-buffer-message))))
	(while x
	  (if (and (buffer-name (car x))
		   (string-match regexp (buffer-name (car x))))
	      (bury-buffer (car x)))
	  (setq x (cdr x))))
    (bury-buffer (mew-buffer-message)))
  (let ((x mew-clean-up-buffers))
    (while x
      (if (get-buffer (car x))
	  (bury-buffer (car x)))
      (setq x (cdr x))))
  (set-buffer (car (buffer-list)))
  (run-hooks 'mew-suspend-hook)
  )

(defun mew-kill-buffer (&optional buf)
  "Erase the current mode(buffer)."
  (interactive)
  (if buf
      (if (get-buffer buf) (kill-buffer buf))
    (kill-buffer (current-buffer)))
  )

(defun mew-summary-quit ()
  "Quit Mew. All buffers of Mew are erased."
  (interactive)
  (if (not (mew-y-or-n-p "Quit Mew? "))
      ()
    (mew-window-pop)
    (setq mew-mail-path nil)
    (setq mew-news-path nil)
    (setq mew-window-stack nil)
    (setq mew-draft-folder nil)
    (setq mew-alias-alist ())
    (setq mew-folder-list ())
    (setq mew-folder-alist ())
    (setq mew-cache nil)
    ;; killing buffers
    ;; message buffers
    (if (and window-system)
	(let ((x (buffer-list))
	      (regexp (concat "^" mew-buffer-message)))
	  (while x
	    (if (and (buffer-name (car x))
		     (string-match regexp (buffer-name (car x))))
		(mew-kill-buffer (car x)))
	    (setq x (cdr x))
	    ))
      (mew-kill-buffer (mew-buffer-message)))
    ;; cache buffers
    (let ((x (buffer-list))
	  (regex (concat "^" (regexp-quote mew-buffer-cache))))
      (while x
	(if (and (buffer-name (car x))
		 (string-match regex (buffer-name (car x))))
	    (mew-kill-buffer (car x)))
	(setq x (cdr x))
	))
    ;;
    (mew-kill-buffer mew-buffer-hello)
    (mew-kill-buffer mew-buffer-mime)
    (mew-kill-buffer mew-buffer-tmp)
    ;; xxx summary & draft buffers...
    ;;
    (remove-hook 'kill-emacs-hook (function mew-mark-clean-up-all))
    (mew-temp-dir-clean-up)
    (setq mew-temp-dir nil mew-temp-file nil)
    ;;
    (remove-hook 'kill-emacs-hook (function mew-temp-dir-clean-up))
    (mew-mark-clean-up-all)
    (while mew-clean-up-buffers
      (mew-kill-buffer (car mew-clean-up-buffers))
      (setq mew-clean-up-buffers (cdr mew-clean-up-buffers))
      )
    (setq mew-clean-up-buffers (list mew-inbox-folder))
    ;;
    (run-hooks 'mew-quit-hook)
    )
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Load Mew libraries
;;;
(require 'mew-complete)
(require 'mew-minibuf)
(require 'mew-cache)
(require 'mew-encode)
(require 'mew-decode)
(require 'mew-mime)
(require 'mew-mark)
(require 'mew-header)
(require 'mew-pgp)
(require 'mew-bq)
(require 'mew-syntax)
(require 'mew-scan)
(require 'mew-pick)
(require 'mew-summary)
(require 'mew-virtual)
(require 'mew-message)
(require 'mew-draft)
(require 'mew-attach)
(require 'mew-demo)
(require 'mew-refile)
(require 'mew-ext)
(require 'mew-fib)
(require 'mew-sort)
(require 'mew-highlight)

;;;
;;; End of Mew
;;;
